SystemOrganization addCategory: #'Shingle-Snapshot'!
SystemOrganization addCategory: #'Shingle-Copies'!

TestCase subclass: #SGCopyCopy
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Shingle-Copies'!

!SGCopyCopy methodsFor: 'running' stamp: 'lr 2/16/2007 14:24'!
tearDown
	Smalltalk at: #SGMock ifPresent: [ :class | class removeFromSystem ]! !

!SGCopyCopy methodsFor: 'testing' stamp: 'lr 2/16/2007 14:25'!
testClass
	| class snapshot1 snapshot2 snapshot3 |
	class := Object subclass: #SGMock
		instanceVariableNames: ''
		classVariableNames: ''
		poolDictionaries: ''
		category: self class category.

	snapshot1 := class snapshotCopy.
	self assert: (class isIdenticalToSnapshot: snapshot1).
	
	class addInstVarName: 'foo'.
	snapshot2 := class snapshotCopy.
	self assert: (class isIdenticalToSnapshot: snapshot2).
	
	class compile: 'foo ^ foo'.
	snapshot3 := class snapshotCopy.
	self assert: (class isIdenticalToSnapshot: snapshot3).

	snapshot1 restore: class.
	self assert: class instVarNames isEmpty.
	self assert: class methodDict isEmpty.
	
	snapshot2 restore: class.
	self assert: class instVarNames = #('foo').
	self assert: class methodDict isEmpty.
	
	snapshot3 restore: class.
	self assert: class instVarNames = #('foo').
	self assert: class methodDict notEmpty! !

!SGCopyCopy methodsFor: 'testing' stamp: 'lr 2/16/2007 14:21'!
testIndexed
	| a1 s1 a2 |
	a1 := #( 1 2 3 ) copy.
	s1 := a1 snapshotCopy.
	self assert: (a1 isIdenticalToSnapshot: s1).
	
	a2 := #( 3 2 1 ) copy.
	self deny: (a2 isIdenticalToSnapshot: s1).
	a2 restoreFromSnapshot: s1.
	self assert: (a2 isIdenticalToSnapshot: s1).
	
	self assert: a1 = a2! !

!SGCopyCopy methodsFor: 'testing' stamp: 'lr 2/16/2007 14:19'!
testNamed
	| p1 s1 p2 |
	p1 := 1 @ 2.
	s1 := p1 snapshotCopy.
	self assert: (p1 isIdenticalToSnapshot: s1).
	
	p2 := Point new.
	self deny: (p2 isIdenticalToSnapshot: s1).
	p2 restoreFromSnapshot: s1.
	self assert: (p2 isIdenticalToSnapshot: s1).
	
	self assert: p1 = p2! !

TestCase subclass: #SGTestSnapshot
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Shingle-Snapshot'!

!SGTestSnapshot methodsFor: 'testing' stamp: 'lr 2/9/2007 17:18'!
testAncestors
	| object step0 step1 step21 step22 |
	step0 := SGSnapshot new.
	step0 register: (object := ValueHolder new).

	object contents: 1.
	step1 := step0 snapshot.
	
	step1 activate.
	object contents: 21.
	step21 := step1 snapshot.
	
	step1 activate.
	object contents: 22.
	step22 := step1 snapshot.
	
	self assert: (step21 ancestors includes: step1).
	self assert: (step21 allAncestors includes: step0).
	self assert: (step21 allAncestors includes: step1).
	
	self assert: (step22 ancestors includes: step1).
	self assert: (step22 allAncestors includes: step0).
	self assert: (step22 allAncestors includes: step1)
! !

!SGTestSnapshot methodsFor: 'testing' stamp: 'lr 2/9/2007 17:19'!
testArray
	| array step0 step1 step2 |
	step0 := SGSnapshot new.
	step0 register: (array := Array new: 1).
	
	array at: 1 put: 1.
	step1 := step0 snapshot.
	array at: 1 put: 2.
	step2 := step1 snapshot.
	
	step0 activate.
	self assert: array first isNil.
	step1 activate.
	self assert: array first = 1.
	step2 activate.
	self assert: array first = 2! !

!SGTestSnapshot methodsFor: 'testing-merging' stamp: 'lr 2/9/2007 17:25'!
testMergeConflict
	| object step0 step11 step12 step21 step22 |
	step0 := SGSnapshot new.
	step0
		register: (object := Array new: 2);
		register: (object at: 1 put: ValueHolder new);
		register: (object at: 2 put: ValueHolder new);
		update.

	step0 activate.
	object first contents: 1.
	step11 := step0 snapshot.
	
	step0 activate.
	object first contents: 2.
	object second contents: 3.
	step12 := step0 snapshot.
	
	self 
		should: [ step11 merge: step12 ]
		raise: Error.
	self 
		should: [ step12 merge: step11 ]
		raise: Error.

	step21 := step11 merge: step12 resolver: [ :a :b |
		self assert: a contents = 1; assert: b contents = 2.
		ValueHolder new contents: 4 ].
	step21 activate.
  	self assert: object first contents = 4.
	self assert: object second contents = 3.
	
	step22 := step12 merge: step11 resolver: [ :a :b |
		self assert: a contents = 2; assert: b contents = 1.
		ValueHolder new contents: 5 ].
	step22 activate.
  	self assert: object first contents = 5.
	self assert: object second contents = 3! !

!SGTestSnapshot methodsFor: 'testing-merging' stamp: 'lr 2/9/2007 17:27'!
testMergeSimple
	| object step0 step11 step12 step21 step22 |
	step0 := SGSnapshot new.
	step0
		register: (object := Array new: 2);
		register: (object at: 1 put: ValueHolder new);
		register: (object at: 2 put: ValueHolder new);
		update.

	step0 activate.
	object first contents: 1.
	step11 := step0 snapshot.
	
	step0 activate.
	object second contents: 2.
	step12 := step0 snapshot.
	
	step21 := step11 merge: step12.
	step21 activate.
  	self assert: object first contents = 1.
	self assert: object second contents = 2.
	
	step22 := step12 merge: step11.
	step22 activate.
  	self assert: object first contents = 1.
	self assert: object second contents = 2! !

!SGTestSnapshot methodsFor: 'testing' stamp: 'lr 2/16/2007 13:45'!
testObject
	| object step0 step1 step2 |
	step0 := SGSnapshot new.
	step0 register: (object := ValueHolder new).

	object contents: 1.
	step1 := step0 snapshot.
	object contents: 2.
	step2 := step1 snapshot.
	
	step0 activate.
	self assert: object contents isNil.
	step1 activate.
	self assert: object contents = 1.
	step2 activate.
	self assert: object contents = 2! !

!SGTestSnapshot methodsFor: 'testing' stamp: 'lr 2/9/2007 17:47'!
testTransaction
	| object step0 |
	step0 := SGSnapshot new.
	step0 register: (object := ValueHolder new).
	step0 transaction: [ object contents: 1 ].
	
	self assert: object contents = 1.
	self
		should: [ step0 transaction: [ object contents: 2. 1 / 0 ] ]
		raise: ZeroDivide.
	self assert: object contents = 1! !

!Object methodsFor: '*shingle' stamp: 'lr 2/16/2007 14:03'!
isIdenticalToSnapshot: aCopy
	^ aCopy isIdenticalTo: self! !

!Object methodsFor: '*shingle' stamp: 'lr 2/12/2007 09:57'!
restoreFromSnapshot: aCopy
	aCopy restore: self! !

!Object methodsFor: '*shingle' stamp: 'lr 2/12/2007 09:56'!
snapshotCopy
	^ SGObjectCopy from: self! !

Object subclass: #SGCopy
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Shingle-Copies'!

SGCopy subclass: #SGClassCopy
	instanceVariableNames: 'name environment type category superclass varNames classVarNames methods classMethods'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Shingle-Copies'!

!SGClassCopy class methodsFor: 'instance-creation' stamp: 'lr 2/12/2007 10:20'!
from: aClass
	^ self basicNew snapshot: aClass! !

!SGClassCopy methodsFor: 'testing' stamp: 'lr 2/16/2007 14:05'!
isIdenticalTo: aClass
	^ name = aClass name
		and: [ environment = aClass environment
		and: [ type = aClass typeOfClass
		and: [ category = aClass category
		and: [ superclass = aClass superclass
		and: [ varNames = aClass instVarNames
		and: [ classVarNames = aClass class instVarNames
		and: [ methods = aClass methodDict 
		and: [ classMethods = aClass class methodDict ] ] ] ] ] ] ] ]! !

!SGClassCopy methodsFor: 'accessing' stamp: 'lr 2/16/2007 11:22'!
priority
	^ -10! !

!SGClassCopy methodsFor: 'public' stamp: 'lr 2/12/2007 16:53'!
restore: aClass
	aClass methodDict: methods.
	aClass class methodDict: classMethods.
	ClassBuilder beSilentDuring: [
		ClassBuilder new
			class: aClass
			name: name
			inEnvironment: environment
			subclassOf: superclass
			type: type
			instanceVariableNames: varNames
			classVariableNames: classVarNames
			category: category ]! !

!SGClassCopy methodsFor: 'public' stamp: 'lr 2/12/2007 16:53'!
snapshot: aClass
	name := aClass name.
	environment := aClass environment.
	type := aClass typeOfClass.
	category := aClass category.
	superclass := aClass superclass.
	varNames := aClass instVarNames.
	classVarNames := aClass class instVarNames.
	methods := aClass methodDict copy.
	classMethods := aClass class methodDict copy! !

!SGCopy class methodsFor: 'instance-creation' stamp: 'lr 2/12/2007 09:28'!
from: anObject
	"Intern the state of anObject."
	
	self subclassResponsibility! !

!SGCopy methodsFor: 'testing' stamp: 'lr 2/16/2007 14:03'!
isIdenticalTo: anObject
	^ false! !

!SGCopy methodsFor: 'accessing' stamp: 'lr 2/16/2007 11:22'!
priority
	self subclassResponsibility ! !

!SGCopy methodsFor: 'public' stamp: 'lr 2/12/2007 08:36'!
restore: anObject
	"Restore the interned state to anObject."
	
	self subclassResponsibility! !

!SGCopy methodsFor: 'public' stamp: 'lr 2/12/2007 10:21'!
snapshot: anObject
	"Intern the state of anObject."
	
	self subclassResponsibility! !

SGCopy variableSubclass: #SGObjectCopy
	instanceVariableNames: 'namedSize'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Shingle-Copies'!

!SGObjectCopy commentStamp: 'lr 2/12/2007 09:22' prior: 0!
I intern the state of a Squeak object with named and indexed variables in space efficient manner. I am independent of the original object, so even if its class is gone I do not loose any information. 

My internal structure is as follows:

	| name_1 | name_2 | ... | name_n 
	| var_1 | var_2 | ... | var_n 
	| indexed_1 | indexed_2 | ... | indexed_m
	
where n is the number of named variables and m is the number of indexed variables.!

!SGObjectCopy class methodsFor: 'instance-creation' stamp: 'lr 2/12/2007 09:54'!
from: anObject
	^ (self basicNew: 2 * anObject class instSize + anObject basicSize) snapshot: anObject! !

!SGObjectCopy methodsFor: 'querying' stamp: 'lr 2/12/2007 09:27'!
indexedAt: anInteger ifAbsent: aBlock
	"Answer the value of an indexed variable."
	
	| index |
	index := 2 * namedSize + anInteger.
	^ (0 < anInteger and: [ index <= self basicSize ])
		ifTrue: [ self basicAt: index ]
		ifFalse: [ aBlock value ]! !

!SGObjectCopy methodsFor: 'accessing' stamp: 'lr 2/12/2007 09:17'!
indexedSize
	^ self basicSize - (2 * self namedSize)! !

!SGObjectCopy methodsFor: 'initialization' stamp: 'lr 2/12/2007 09:47'!
initializeFrom: anObject
	| names index |
	names := anObject class allInstVarNames.
	namedSize := index := names size.
	[ index > 0 ] whileTrue: [
		self 
			basicAt: index 
			put: (names at: index).
		self
			basicAt: index + namedSize 
			put: (anObject instVarAt: index).
		index := index - 1 ].
	index := anObject basicSize.
	[ index > 0 ] whileTrue: [
		self 
			basicAt: 2 * namedSize + index 
			put: (anObject basicAt: index).
		index := index - 1 ]! !

!SGObjectCopy methodsFor: 'testing' stamp: 'lr 2/16/2007 14:17'!
isIdenticalTo: anObject
	| index names |
	index := self namedSize.
	names := anObject class allInstVarNames.
	index = names size 
		ifFalse: [ ^ false ].
	[ index > 0 ] whileTrue: [
		(names at: index) = (self basicAt: index)
			ifFalse: [ ^ false ].
		(anObject instVarAt: index) == (self basicAt: namedSize + index)
			ifFalse: [ ^ false ].
		index := index - 1 ].
	index := self indexedSize.
	index = anObject basicSize
		ifFalse: [ ^ false ].
	[ index > 0 ] whileTrue: [
		(anObject basicAt: index) == (self basicAt: 2 * namedSize + index)
			ifFalse: [ ^ false ].
		index := index - 1 ].
	^ true! !

!SGObjectCopy methodsFor: 'querying' stamp: 'lr 2/12/2007 09:24'!
namedAt: aString ifAbsent: aBlock
	"Answer the value of a named variable."
	
	| index |
	index := namedSize.
	[ index > 0 ] whileTrue: [
		(self basicAt: index) = aString
			ifTrue: [ ^ self basicAt: namedSize + index ].
		index := index - 1 ].
	^ aBlock value! !

!SGObjectCopy methodsFor: 'accessing' stamp: 'lr 2/12/2007 09:12'!
namedSize
	^ namedSize! !

!SGObjectCopy methodsFor: 'accessing' stamp: 'lr 2/16/2007 11:22'!
priority
	^ 10! !

!SGObjectCopy methodsFor: 'public' stamp: 'lr 2/16/2007 13:47'!
restore: anObject
	| names index |
	names := anObject class allInstVarNames.
	index := names size.
	[ index > 0 ] whileTrue: [
		anObject
			instVarAt: index 
			put: (self 
				namedAt: (names at: index)
				ifAbsent: [ nil ]).
		index := index - 1 ].
	index := anObject basicSize.
	[ index > 0 ] whileTrue: [
		anObject 
			basicAt: index 
			put: (self 
				indexedAt: index
				ifAbsent: [ nil ]).
		index := index - 1 ]! !

!SGObjectCopy methodsFor: 'public' stamp: 'lr 2/12/2007 09:49'!
snapshot: anObject
	| names index |
	names := anObject class allInstVarNames.
	namedSize := index := names size.
	[ index > 0 ] whileTrue: [
		self 
			basicAt: index 
			put: (names at: index).
		self
			basicAt: index + namedSize 
			put: (anObject instVarAt: index).
		index := index - 1 ].
	index := anObject basicSize.
	[ index > 0 ] whileTrue: [
		self 
			basicAt: 2 * namedSize + index 
			put: (anObject basicAt: index).
		index := index - 1 ]! !

Object subclass: #SGSnapshot
	instanceVariableNames: 'objects ancestors'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Shingle-Snapshot'!

!SGSnapshot methodsFor: 'actions' stamp: 'lr 2/16/2007 13:57'!
activate
	"Activate the receiving snapshot."
	
	self objectsAndCopiesDo: [ :object :copy | 
		object restoreFromSnapshot: copy ]! !

!SGSnapshot methodsFor: 'ancestors' stamp: 'lr 2/5/2007 10:22'!
addAncestor: aSnapshot
	ancestors := ancestors copyWith: aSnapshot! !

!SGSnapshot methodsFor: 'ancestors' stamp: 'lr 2/5/2007 11:02'!
allAncestors
	^ Array streamContents: [ :stream | 
		self allAncestorsDo: [ :each |
			stream nextPut: each ] ]! !

!SGSnapshot methodsFor: 'ancestors' stamp: 'lr 2/5/2007 10:22'!
allAncestorsDo: aBlock
	| seen todo next |
	seen := Set with: self.
	todo := OrderedCollection with: self.
	[ todo isEmpty ] whileFalse: [
		next := todo removeFirst.
		next ancestors do: [ :each |
			(seen includes: each) ifFalse: [
				aBlock value: each.
				seen add: each.
				todo add: each ] ] ]! !

!SGSnapshot methodsFor: 'accessing' stamp: 'lr 2/5/2007 09:55'!
ancestors
	^ ancestors! !

!SGSnapshot methodsFor: 'enumerating' stamp: 'lr 2/16/2007 14:29'!
at: anObject ifAbsent: aBlock
	self objectsAndCopiesDo: [ :object :copy |
		anObject == object
			ifTrue: [ ^ copy ] ].
	^ aBlock value! !

!SGSnapshot methodsFor: 'ancestors' stamp: 'lr 2/5/2007 12:12'!
commonAncestorsWith: aSnapshot
	^ self allAncestors intersection: aSnapshot allAncestors! !

!SGSnapshot methodsFor: 'enumerating' stamp: 'lr 2/16/2007 13:59'!
indexObjectsAndCopiesDo: aBlock
	1 to: objects size by: 2 do: [ :index |
		aBlock
			value: index
			value: (objects at: index) 
			value: (objects at: index + 1) ]! !

!SGSnapshot methodsFor: 'initialization' stamp: 'lr 2/16/2007 14:00'!
initialize
	super initialize.
	objects := ancestors := #()! !

!SGSnapshot methodsFor: 'public' stamp: 'lr 2/16/2007 14:45'!
merge: aSnapshot
	^ self merge: aSnapshot resolver: [ :obj :a :b | self error: 'Merge conflict' ]! !

!SGSnapshot methodsFor: 'public' stamp: 'lr 2/16/2007 14:48'!
merge: aSnapshot resolver: aBlock
	| result ancestor common other |
	result := self class new
		addAncestor: self;
		addAncestor: aSnapshot;
		yourself.
	ancestor := self commonAncestorsWith: aSnapshot.
	ancestor := ancestors isEmpty
		ifTrue: [ self class new ]
		ifFalse: [ ancestors first ].
	self objectsAndCopiesDo: [ :object :copy |
		common := ancestor at: object ifAbsent: [ nil ].
		other := aSnapshot at: object ifAbsent: [ nil ].
		(copy == common
			ifTrue: [ result register: object with: other ]
			ifFalse: [
				other == common
					ifTrue: [ result register: object with: copy ] ]) ].
	aSnapshot objectsAndCopiesDo: [ :object :copy |
		common := ancestor at: object ifAbsent: [ nil ].
		other := self at: object ifAbsent: [ nil ].
		result register: object with: (copy == common
			ifTrue: [ other ]
			ifFalse: [
				other == common
					ifTrue: [ copy ]
					ifFalse: [ 
						(aBlock 
							value: (object shallowCopy
								restoreFromSnapshot: other)
							value: (object shallowCopy
								restoreFromSnapshot: copy))
									snapshotCopy ] ]) ].
	^ result! !

!SGSnapshot methodsFor: 'accessing' stamp: 'lr 2/5/2007 09:55'!
objects
	^ objects! !

!SGSnapshot methodsFor: 'enumerating' stamp: 'lr 2/16/2007 13:57'!
objectsAndCopiesDo: aBlock
	1 to: objects size by: 2 do: [ :index |
		aBlock 
			value: (objects at: index) 
			value: (objects at: index + 1) ]! !

!SGSnapshot methodsFor: 'copying' stamp: 'lr 2/16/2007 13:48'!
postCopy
	super postCopy.
	objects := objects copy.
	ancestors := #()! !

!SGSnapshot methodsFor: 'printing' stamp: 'lr 2/5/2007 12:17'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: ' <'; print: self identityHash; nextPutAll: '>'! !

!SGSnapshot methodsFor: 'public' stamp: 'lr 2/16/2007 14:30'!
register: anObject
	self register: anObject with: anObject snapshotCopy! !

!SGSnapshot methodsFor: 'public' stamp: 'lr 2/16/2007 14:30'!
register: anObject with: aCopy
	| pair |
	pair := Array with: anObject with: aCopy.
	objects := aCopy priority < 0 ifTrue: [ pair , objects ] ifFalse: [ objects , pair ]! !

!SGSnapshot methodsFor: 'actions' stamp: 'lr 2/9/2007 17:15'!
snapshot
	"Answer a new snapshot intering the current state of all registered objects."
	
	^ self copy
		addAncestor: self;
		update! !

!SGSnapshot methodsFor: 'utilities' stamp: 'lr 2/9/2007 17:46'!
transaction: aBlock
	| snapshot |
	snapshot := self snapshot.
	aBlock ifCurtailed: [ snapshot activate ]! !

!SGSnapshot methodsFor: 'actions' stamp: 'lr 2/16/2007 14:00'!
update
	"Updates the current snapshot from the real objects."
	
	self indexObjectsAndCopiesDo: [ :index :object :copy |
		(object isIdenticalToSnapshot: copy)
			ifFalse: [ objects basicAt: index + 1 put: object snapshotCopy ] ]! !

!Class methodsFor: '*shingle' stamp: 'lr 2/12/2007 16:44'!
snapshotCopy
	^ SGClassCopy from: self! !

!ClassBuilder methodsFor: '*shingle' stamp: 'lr 2/12/2007 16:54'!
class: aClass name: aNameString inEnvironment: anEnvironment subclassOf: aSuperClass type: aTypeSymbol instanceVariableNames: anInstVarString classVariableNames: aClassVarString category: aCategoryString
	| instVars oldClass copyOfOldClass needNew newClass force newCategory organization |
	instVars := Scanner new scanFieldNames: anInstVarString.
	oldClass := aClass.
	oldClass isBehavior ifTrue: [
		copyOfOldClass := oldClass copy.
		copyOfOldClass superclass addSubclass: copyOfOldClass ].
	[	needNew := self needsSubclassOf: aSuperClass type: aTypeSymbol instanceVariables: instVars from: oldClass.
		needNew ifNil: [ ^ nil ].
		needNew 
			ifFalse: [ newClass := oldClass ]
			ifTrue: [
				newClass := self 
					newSubclassOf: aSuperClass 
					type: aTypeSymbol 
					instanceVariables: instVars
					from: oldClass.
				newClass ifNil: [ ^ nil ].
				newClass setName: aNameString ].
		force := newClass declare: aClassVarString.
		newCategory := aCategoryString asSymbol.
		organization := anEnvironment ifNotNil: [ anEnvironment organization ].
		organization classify: newClass name under: newCategory.
		newClass environment: anEnvironment.
		newClass := self recompile: force from: oldClass to: newClass mutate: false.
		(anEnvironment at: newClass name ifAbsent: [ nil ]) == newClass ifFalse: [
			[ environ at: newClass name put: newClass ]
				on: AttemptToWriteReadOnlyGlobal
				do: [ :ex | ex resume: true ].
			Smalltalk flushClassNameCache ].
		newClass doneCompiling ] 
	ensure: [
		copyOfOldClass ifNotNil: [ copyOfOldClass superclass removeSubclass: copyOfOldClass ].
		Behavior flushObsoleteSubclasses ].
	^ newClass! !

!ValueHolder methodsFor: '*shingle-snapshot' stamp: 'lr 2/5/2007 14:44'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: ' print: '; print: self contents! !
